@@ -1,12 +1,15 @@
 # from scapy.all import *
 from scapy.layers.inet import TCP, IP, UDP
 from scapy.sendrecv import sniff
-from threading import Thread
+from scapy.utils import PcapReader
+from joblib import Parallel, delayed
+import signal
+from lockfile import LockFile
 import pandas as pd
 import numpy as np
 import binascii
 import gc
-from .flowsession import FlowSession
+import itertools
 
 class Flowmeter:
 
@@ -17,7 +20,7 @@ class Flowmeter:
     learning applications.
     """
     
-    def __init__(self, offline=None, outfunc=None, outfile=None):
+    def __init__(self, offline=None, outfunc=None, outfile=None, buffsize=1024):
         
         """
         Args:
@@ -30,6 +33,8 @@ class Flowmeter:
         self.outfunc = outfunc
         self.offline = offline
         self.outfile = outfile
+        self.buffsize=buffsize
+        self._sess_frames = dict()
         self.columns = [
             "flow",                 # Index
             "src",                  # Source IP
@@ -145,24 +150,41 @@ class Flowmeter:
                 f.write(','.join(self.columns) + '\n')
 
     
-    def process_session(self, packetlist):
-        self._pcap = packetlist
-        out_string = self._build_feature_dataframe().to_csv(header=False, line_terminator='\n')
+    def process_session(self, session_df):
+        for column in session_df.columns:
+            session_df[column] = session_df[column].replace(r'\s+', np.nan, regex=True)
+            session_df[column] = session_df[column].fillna(0)
+        out_string = session_df.to_csv(header=False, index=False, line_terminator='\n')
         self.outfunc(out_string)
         if self.outfile is not None:
+            lock = LockFile(self.outfile)
+            lock.acquire()
             with open(self.outfile, 'at', newline='') as f:
                 f.write(out_string)
-
+            lock.release()
 
     def run(self):
         session_kwargs = dict(
             callback=self.process_session,
             sess_function=self._get_sessions
         )
-        print('Sniffing; ctrl+c to stop.')
-        results = sniff(session=FlowSession, offline=self.offline, session_kwargs=session_kwargs)
-        print('Stopping...')
-        self.process_session(results)
+
+        if self.offline:
+            reader = PcapReader(self.offline)
+            Parallel(-1)(delayed(self.process_packet)(packet) for packet in reader)
+
+        else:
+            print('Sniffing; ctrl+c to stop.') 
+            results = sniff(prn=self.process_packet)
+            print('Stopping...')
+            for packet in results:
+                self.process_packet
+        
+        
+    def empty(self):
+        keys = list(self._sess_frames.keys())
+        for flow in keys:
+            self.process_session(self._build_feature_from_flow(self._sess_frames.pop(flow)))
         
 
     def _get_sessions(self, packet):
@@ -273,6 +295,55 @@ class Flowmeter:
         df = df.drop(columns="index")
         return df
 
+    def build_dataframe_row(self, packet):
+
+        """
+        This function takes in a scapy PacketList object and 
+        builds a pandas dataframe.
+
+        Args:
+            packet_list (PacketList): A scapy PacketList object.
+        
+        """
+        ip_fields = [field.name for field in IP().fields_desc]
+        tcp_fields = [field.name for field in TCP().fields_desc]
+        # udp_fields = [field.name for field in UDP().fields_desc if field.name not in tcp_fields]
+
+        dataframe_fields = ip_fields + ['time', 'protocol'] + tcp_fields + ['size','payload','payload_raw','payload_hex'] 
+
+        # Create blank DataFrame
+        # df = pd.DataFrame(columns=dataframe_fields)
+        # Field array for each row of DataFrame
+        field_values = []
+        # Add all IP fields to dataframe
+        for field in ip_fields:
+            if field == 'options':
+                # Retrieving number of options defined in IP Header
+                field_values.append(len(packet[IP].fields[field]))
+            else:
+                field_values.append(packet[IP].fields[field])
+
+        field_values.append(packet.time)
+        layer_type = type(packet[IP].payload)
+
+        field_values.append(layer_type().name)
+        for field in tcp_fields:
+            try:
+                if field == 'options':
+                    field_values.append(len(packet[layer_type].fields[field]))
+                else:
+                    field_values.append(packet[layer_type].fields[field])
+                
+            except:
+                field_values.append(None)
+        
+        # Append payload
+        field_values.append(len(packet))
+        field_values.append(len(packet[layer_type].payload))
+        field_values.append(packet[layer_type].payload.original)
+        field_values.append(binascii.hexlify(packet[layer_type].payload.original))
+        return dataframe_fields, field_values
+
     def build_sessions(self):
 			  
         """
@@ -1482,136 +1553,132 @@ class Flowmeter:
             return ("{}:{}<->{}:{}").format(src,str(src_port[0]),dst,str(src_port[0]))
 
  
-    def _build_feature_from_flow(self, flow):
+    def _build_feature_from_flow(self, flow_df):
 
         gc.collect()
         # print(("\nEntering {}").format(flow)) # Test
 
-        if flow == "Other" or "Ethernet" in flow or "ARP" in flow: 
-            pass
+        result = pd.DataFrame(columns=self.columns)
+        result["flow"] = [self.build_index(flow_df)]
+        result["src"] = [self.get_src_ip(flow_df)]
+        result["src_port"] = [self.get_src_port(flow_df)]
+        result["dst"] = [self.get_dst_ip(flow_df)]
+        result["dst_port"] = [self.get_dst_port(flow_df)]
+        result["protocol"] = [self.get_protocol(flow_df)]
+        result['timestamp'] = [self.get_timestamp(flow_df)]
+        result["duration"] = [self.get_flow_duration(flow_df)]
+        
+        result["total_fpackets"] = [self.get_total_forward_packets(flow_df)]
+        result["total_bpackets"] = [self.get_total_backward_packets(flow_df)]
+        result["total_fpktl"] = [self.get_total_len_forward_packets(flow_df)]
+        result["total_bpktl"] = [self.get_total_len_backward_packets(flow_df)]
+        
+        result["min_fpktl"] = [self.get_min_forward_packet_size(flow_df)]
+        result["max_fpktl"] = [self.get_max_forward_packet_size(flow_df)]
+        result["mean_fpktl"] = [self.get_mean_forward_packet_size(flow_df)]
+        result["std_fpktl"] = [self.get_std_forward_packet_size(flow_df)]
+        
+        result["min_bpktl"] = [self.get_min_backward_packet_size(flow_df)]
+        result["max_bpktl"] = [self.get_max_backward_packet_size(flow_df)]
+        result["mean_bpktl"] = [self.get_mean_backward_packet_size(flow_df)]
+        result["std_bpktl"] = [self.get_std_backward_packet_size(flow_df)]
+        
+        result["flowBytesPerSecond"] = [self.get_flow_bytes_per_second(flow_df)]
+        result["flowPktsPerSecond"] = [self.get_flow_packets_per_second(flow_df)]
+        
+        result["mean_flowiat"] = [self.get_mean_flow_iat(flow_df)]
+        result["std_flowiat"] = [self.get_std_flow_iat(flow_df)]
+        result["max_flowiat"] = [self.get_max_flow_iat(flow_df)]
+        result["min_flowiat"] = [self.get_min_flow_iat(flow_df)]
+        
+        result["total_fiat"] = [self.get_iat_forward_total_time(flow_df)]
+        result["mean_fiat"] = [self.get_iat_forward_mean_times(flow_df)]
+        result["std_fiat"] = [self.get_iat_forward_std_times(flow_df)]
+        result["max_fiat"] = [self.get_iat_forward_max_times(flow_df)]
+        result["min_fiat"] = [self.get_iat_forward_min_times(flow_df)]
+        
+        result["total_biat"] = [self.get_iat_backward_total_time(flow_df)]
+        result["mean_biat"] = [self.get_iat_backwards_mean_times(flow_df)]
+        result["std_biat"] = [self.get_iat_backwards_std_times(flow_df)]
+        result["max_biat"] = [self.get_iat_forward_max_times(flow_df)]
+        result["min_biat"] = [self.get_iat_backwards_min_times(flow_df)]
+        
+        result["fpsh_cnt"] = [self.get_total_forward_push_flags(flow_df)]
+        result["bpsh_cnt"] = [self.get_total_backward_push_flags(flow_df)]
+        
+        result["furg_cnt"] = [self.get_total_forward_urgent_flags(flow_df)]
+        result["burg_cnt"] = [self.get_total_backward_urgent_flags(flow_df)]
+        
+        result["total_fhlen"] = [self.get_total_header_len_forward_packets(flow_df)]
+        result["total_bhlen"] = [self.get_total_header_len_backward_packets(flow_df)]
+        
+        result["fPktsPerSecond"] = [self.get_forward_packets_per_second(flow_df)]
+        result["bPktsPerSecond"] = [self.get_backward_packets_per_second(flow_df)]
+        
+        result["min_flowpktl"] = [self.get_min_flow_packet_size(flow_df)]
+        result["max_flowpktl"] = [self.get_max_flow_packet_size(flow_df)]
+        result["mean_flowpktl"] = [self.get_mean_flow_packet_size(flow_df)]
+        result["std_flowpktl"] = [self.get_std_flow_packet_size(flow_df)]
+        result["var_flowpktl"] = [self.get_var_flow_packet_size(flow_df)]
+        
+        result["flow_fin"] = [self.get_total_flow_fin_flags(flow_df)]
+        result["flow_syn"] = [self.get_total_flow_syn_flags(flow_df)]
+        result["flow_rst"] = [self.get_total_flow_reset_flags(flow_df)]
+        result["flow_psh"] = [self.get_total_flow_push_flags(flow_df)]
+        result["flow_ack"] = [self.get_total_flow_ack_flags(flow_df)]
+        result["flow_urg"] = [self.get_total_flow_urg_flags(flow_df)]
+        result["flow_cwr"] = [self.get_total_flow_cwr_flags(flow_df)]
+        result["flow_ece"] = [self.get_total_flow_ece_flags(flow_df)]
+        
+        result["downUpRatio"] = [self.get_upload_download_ratio(flow_df)]
+        
+        result["avgPacketSize"] = [self.get_avg_packet_size(flow_df)]
+        result["fAvgSegmentSize"] = [self.get_avg_forward_segment_size(flow_df)]
+        result["bAvgSegmentSize"] = [self.get_avg_backward_segment_size(flow_df)]
+        
+        result["fAvgBytesPerBulk"] = [self.get_average_forward_bytes_per_burt(flow_df)]
+        result["fAvgPacketsPerBulk"] = [self.get_avg_forward_burst_packets(flow_df)]
+        result["fAvgBulkRate"] = [self.get_avg_forward_in_total_burst(flow_df)]
+        
+        result["bAvgBytesPerBulk"] = [self.get_average_backward_bytes_per_burt(flow_df)]
+        result["bAvgPacketsPerBulk"] = [self.get_avg_backward_burst_packets(flow_df)]
+        result["bAvgBulkRate"] = [self.get_avg_backward_in_total_burst(flow_df)]
+        
+        result["fSubFlowAvgPkts"] = [self.get_fwd_subflow_packets(flow_df)]
+        result["fSubFlowAvgBytes"] = [self.get_fwd_subflow_bytes(flow_df)]
+        result["bSubFlowAvgPkts"] = [self.get_bwd_subflow_packets(flow_df)]
+        result["bSubFlowAvgBytes"] = [self.get_bwd_subflow_bytes(flow_df)]
+        
+        result['fInitWinSize'] = [self.get_fwd_init_tcp_win(flow_df)]
+        result['bInitWinSize'] = [self.get_bwd_init_tcp_win(flow_df)]
+        result['fDataPkts'] = [self.get_fwd_data_packets(flow_df)]
+        result['fHeaderSizeMin'] = [self.get_fwd_header_min(flow_df)]
+        active_periods, idle_periods = self.get_period_lengths(flow_df, idle_threshold_s=5)
+        if len(active_periods) > 0:
+            result['mean_active_s'] = [active_periods.mean()]
+            result['std_active_s'] = [active_periods.std()]
+            result['max_active_s'] = [active_periods.max()]
+            result['min_active_s'] = [active_periods.std()]
         else:
-            flow = self.build_dataframe(self._sessions[flow])
-            result = pd.DataFrame(columns=self.columns)
-            result["flow"] = [self.build_index(flow)]
-            result["src"] = [self.get_src_ip(flow)]
-            result["src_port"] = [self.get_src_port(flow)]
-            result["dst"] = [self.get_dst_ip(flow)]
-            result["dst_port"] = [self.get_dst_port(flow)]
-            result["protocol"] = [self.get_protocol(flow)]
-            result['timestamp'] = [self.get_timestamp(flow)]
-            result["duration"] = [self.get_flow_duration(flow)]
+            result['mean_active_s'] = [0]
+            result['std_active_s'] = [0]
+            result['max_active_s'] = [0]
+            result['min_active_s'] = [0]
+        if len(idle_periods) > 0:
+            result['mean_idle_s'] = [idle_periods.mean()]
+            result['std_idle_s'] = [idle_periods.std()]
+            result['max_idle_s'] = [idle_periods.max()]
+            result['min_idle_s'] = [idle_periods.std()]
+        else:
+            result['mean_idle_s'] = [0]
+            result['std_idle_s'] = [0]
+            result['max_idle_s'] = [0]
+            result['min_idle_s'] = [0]
             
-            result["total_fpackets"] = [self.get_total_forward_packets(flow)]
-            result["total_bpackets"] = [self.get_total_backward_packets(flow)]
-            result["total_fpktl"] = [self.get_total_len_forward_packets(flow)]
-            result["total_bpktl"] = [self.get_total_len_backward_packets(flow)]
-            
-            result["min_fpktl"] = [self.get_min_forward_packet_size(flow)]
-            result["max_fpktl"] = [self.get_max_forward_packet_size(flow)]
-            result["mean_fpktl"] = [self.get_mean_forward_packet_size(flow)]
-            result["std_fpktl"] = [self.get_std_forward_packet_size(flow)]
-            
-            result["min_bpktl"] = [self.get_min_backward_packet_size(flow)]
-            result["max_bpktl"] = [self.get_max_backward_packet_size(flow)]
-            result["mean_bpktl"] = [self.get_mean_backward_packet_size(flow)]
-            result["std_bpktl"] = [self.get_std_backward_packet_size(flow)]
-            
-            result["flowBytesPerSecond"] = [self.get_flow_bytes_per_second(flow)]
-            result["flowPktsPerSecond"] = [self.get_flow_packets_per_second(flow)]
-            
-            result["mean_flowiat"] = [self.get_mean_flow_iat(flow)]
-            result["std_flowiat"] = [self.get_std_flow_iat(flow)]
-            result["max_flowiat"] = [self.get_max_flow_iat(flow)]
-            result["min_flowiat"] = [self.get_min_flow_iat(flow)]
-            
-            result["total_fiat"] = [self.get_iat_forward_total_time(flow)]
-            result["mean_fiat"] = [self.get_iat_forward_mean_times(flow)]
-            result["std_fiat"] = [self.get_iat_forward_std_times(flow)]
-            result["max_fiat"] = [self.get_iat_forward_max_times(flow)]
-            result["min_fiat"] = [self.get_iat_forward_min_times(flow)]
-            
-            result["total_biat"] = [self.get_iat_backward_total_time(flow)]
-            result["mean_biat"] = [self.get_iat_backwards_mean_times(flow)]
-            result["std_biat"] = [self.get_iat_backwards_std_times(flow)]
-            result["max_biat"] = [self.get_iat_forward_max_times(flow)]
-            result["min_biat"] = [self.get_iat_backwards_min_times(flow)]
-            
-            result["fpsh_cnt"] = [self.get_total_forward_push_flags(flow)]
-            result["bpsh_cnt"] = [self.get_total_backward_push_flags(flow)]
-            
-            result["furg_cnt"] = [self.get_total_forward_urgent_flags(flow)]
-            result["burg_cnt"] = [self.get_total_backward_urgent_flags(flow)]
-            
-            result["total_fhlen"] = [self.get_total_header_len_forward_packets(flow)]
-            result["total_bhlen"] = [self.get_total_header_len_backward_packets(flow)]
-            
-            result["fPktsPerSecond"] = [self.get_forward_packets_per_second(flow)]
-            result["bPktsPerSecond"] = [self.get_backward_packets_per_second(flow)]
-            
-            result["min_flowpktl"] = [self.get_min_flow_packet_size(flow)]
-            result["max_flowpktl"] = [self.get_max_flow_packet_size(flow)]
-            result["mean_flowpktl"] = [self.get_mean_flow_packet_size(flow)]
-            result["std_flowpktl"] = [self.get_std_flow_packet_size(flow)]
-            result["var_flowpktl"] = [self.get_var_flow_packet_size(flow)]
-            
-            result["flow_fin"] = [self.get_total_flow_fin_flags(flow)]
-            result["flow_syn"] = [self.get_total_flow_syn_flags(flow)]
-            result["flow_rst"] = [self.get_total_flow_reset_flags(flow)]
-            result["flow_psh"] = [self.get_total_flow_push_flags(flow)]
-            result["flow_ack"] = [self.get_total_flow_ack_flags(flow)]
-            result["flow_urg"] = [self.get_total_flow_urg_flags(flow)]
-            result["flow_cwr"] = [self.get_total_flow_cwr_flags(flow)]
-            result["flow_ece"] = [self.get_total_flow_ece_flags(flow)]
-            
-            result["downUpRatio"] = [self.get_upload_download_ratio(flow)]
-            
-            result["avgPacketSize"] = [self.get_avg_packet_size(flow)]
-            result["fAvgSegmentSize"] = [self.get_avg_forward_segment_size(flow)]
-            result["bAvgSegmentSize"] = [self.get_avg_backward_segment_size(flow)]
-            
-            result["fAvgBytesPerBulk"] = [self.get_average_forward_bytes_per_burt(flow)]
-            result["fAvgPacketsPerBulk"] = [self.get_avg_forward_burst_packets(flow)]
-            result["fAvgBulkRate"] = [self.get_avg_forward_in_total_burst(flow)]
-            
-            result["bAvgBytesPerBulk"] = [self.get_average_backward_bytes_per_burt(flow)]
-            result["bAvgPacketsPerBulk"] = [self.get_avg_backward_burst_packets(flow)]
-            result["bAvgBulkRate"] = [self.get_avg_backward_in_total_burst(flow)]
-            
-            result["fSubFlowAvgPkts"] = [self.get_fwd_subflow_packets(flow)]
-            result["fSubFlowAvgBytes"] = [self.get_fwd_subflow_bytes(flow)]
-            result["bSubFlowAvgPkts"] = [self.get_bwd_subflow_packets(flow)]
-            result["bSubFlowAvgBytes"] = [self.get_bwd_subflow_bytes(flow)]
-            
-            result['fInitWinSize'] = [self.get_fwd_init_tcp_win(flow)]
-            result['bInitWinSize'] = [self.get_bwd_init_tcp_win(flow)]
-            result['fDataPkts'] = [self.get_fwd_data_packets(flow)]
-            result['fHeaderSizeMin'] = [self.get_fwd_header_min(flow)]
-            active_periods, idle_periods = self.get_period_lengths(flow, idle_threshold_s=5)
-            if len(active_periods) > 0:
-                result['mean_active_s'] = [active_periods.mean()]
-                result['std_active_s'] = [active_periods.std()]
-                result['max_active_s'] = [active_periods.max()]
-                result['min_active_s'] = [active_periods.std()]
-            else:
-                result['mean_active_s'] = [0]
-                result['std_active_s'] = [0]
-                result['max_active_s'] = [0]
-                result['min_active_s'] = [0]
-            if len(idle_periods) > 0:
-                result['mean_idle_s'] = [idle_periods.mean()]
-                result['std_idle_s'] = [idle_periods.std()]
-                result['max_idle_s'] = [idle_periods.max()]
-                result['min_idle_s'] = [idle_periods.std()]
-            else:
-                result['mean_idle_s'] = [0]
-                result['std_idle_s'] = [0]
-                result['max_idle_s'] = [0]
-                result['min_idle_s'] = [0]
-                
-            result["label"] = ["None"]
-            #print(("\nAppending {}\n").format(result))
-            return result
-            #print(self._frames)
+        result["label"] = ["None"]
+        #print(("\nAppending {}\n").format(result))
+        return result
+        #print(self._frames)
 
 
     def _build_sessions(self):
@@ -1619,23 +1686,18 @@ class Flowmeter:
         return self.build_sessions()
 
 
-    def _build_feature_dataframe(self):        
-        
-        
-        self._sessions = self.build_sessions()
-        gc.collect()
-        # print("\nBuilding Pool\n") # Test
-        # pool = Pool()
-
-        # Not sure this matters when you're not working from a file.
-        # Keeping it on for now.
-        self._frames = [self._build_feature_from_flow(session) for session in self._sessions]
-        # print(self._frames) # Test
-        
-        final = pd.concat(self._frames)
-        final.set_index(["flow"], inplace=True)
-        for column in final.columns:
-            final[column] = final[column].replace(r'\s+', np.nan, regex=True)
-            final[column] = final[column].fillna(0)
-
-        return final
+    def process_packet(self, packet):        
+        sess = self._get_sessions(packet)
+        if sess == "Other" or "Ethernet" in sess or "ARP" in sess: 
+            return
+        fields, row = self.build_dataframe_row(packet)
+        if sess not in self._sess_frames.keys():
+            self._sess_frames[sess] = pd.DataFrame([row], columns=fields)
+        else:
+            self._sess_frames[sess] = self._sess_frames[sess].append(dict(zip(fields, row)), ignore_index=True)
+        curr_df = self._sess_frames[sess]
+        if TCP in packet:
+            if packet[TCP].flags.F or packet[TCP].flags.R:
+                self.process_session(self._build_feature_from_flow(self._sess_frames.pop(sess)))
+        elif curr_df['time'].astype(float).max() - curr_df['time'].astype(float).min() > 120000:
+            self.process_session(self._build_feature_from_flow(self._sess_frames.pop(sess)))
